#!C:\ci\flake8_1552361032895\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'flake8==3.7.7','console_scripts','flake8'
__requires__ = 'flake8==3.7.7'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('flake8==3.7.7', 'console_scripts', 'flake8')()
    )
